import numpy as np
import pytest
from biocframe import BiocFrame
from scipy import sparse
from scrnaseq import fetch_dataset, polish_dataset
from singlecellexperiment import SingleCellExperiment

__author__ = "Jayaram Kancherla"
__copyright__ = "Jayaram Kancherla"
__license__ = "MIT"


def generate_matrix():
    mat = np.random.poisson(1, (100, 10))
    return mat


def test_polish_dataset_strips_assay_dimnames():
    mat = generate_matrix()
    row_names = [f"GENE_{i}" for i in range(mat.shape[0])]
    col_names = list("ABCDEFGHIJ")
    sce = SingleCellExperiment(
        assays={"counts": mat},
        row_data=BiocFrame(row_names=row_names),
        column_data=BiocFrame(row_names=col_names),
    )

    y = polish_dataset(sce)
    assert sce.shape == y.shape
    assert sce.get_assay_names() == y.get_assay_names()
    assert y.get_row_names() is not None
    assert y.get_column_names() is not None

    sce.row_data = sce.row_data.set_row_names([name.lower() for name in row_names])
    sce.column_names = sce.column_data.set_row_names(
        [name.lower() for name in col_names]
    )
    y = polish_dataset(sce)
    assert sce.shape == y.shape
    assert sce.get_assay_names() == y.get_assay_names()
    assert y.get_row_names() == sce.get_row_names()
    assert y.get_column_names() == sce.get_column_names()


def test_polish_dataset_strips_reduced_dimension_names():
    mat = generate_matrix()
    row_names = [f"GENE_{i}" for i in range(mat.shape[0])]
    col_names = list("ABCDEFGHIJ")
    pca = np.random.randn(10, 5)
    sce = SingleCellExperiment(
        assays={"counts": mat},
        row_data=BiocFrame(row_names=row_names),
        column_data=BiocFrame(row_names=col_names),
        reduced_dims={"pca": pca},
    )

    y = polish_dataset(sce)
    assert sce.shape == y.shape
    assert sce.get_assay_names() == y.get_assay_names()
    assert y.get_row_names() is not None
    assert y.get_column_names() is not None
    assert sce.get_reduced_dim_names() == y.get_reduced_dim_names()
    assert y.reduced_dims["pca"].shape == pca.shape


def test_polish_dataset_strips_alternative_experiment_names():
    mat = generate_matrix()
    row_names = [f"GENE_{i}" for i in range(mat.shape[0])]
    col_names = list("ABCDEFGHIJ")
    pca = np.random.randn(10, 5)
    altexp = SingleCellExperiment(
        assays={"counts": mat},
        row_data=BiocFrame(row_names=row_names),
        column_data=BiocFrame(row_names=col_names),
    )
    sce = SingleCellExperiment(
        assays={"counts": mat},
        row_data=BiocFrame(row_names=row_names),
        column_data=BiocFrame(row_names=col_names),
        reduced_dims={"pca": pca},
        alternative_experiments={"rando": altexp},
    )

    y = polish_dataset(sce)
    assert y.alternative_experiments["rando"].shape == altexp.shape
    assert y.alternative_experiments["rando"].get_row_names() == altexp.get_row_names()


def test_polish_dataset_converts_dense_to_sparse():
    mat = np.random.poisson(0.2, (100, 10))
    sce = SingleCellExperiment(assays={"counts": mat})

    y = polish_dataset(sce)
    assert sparse.issparse(y.assays["counts"])

    y = polish_dataset(sce, reformat_assay_by_density=0)
    assert not sparse.issparse(y.assays["counts"])
    assert y.shape == mat.shape

    y = polish_dataset(sce, reformat_assay_by_density=None)
    assert not sparse.issparse(y.assays["counts"])


def test_polish_dataset_converts_sparse_to_dense():
    mat = np.random.poisson(3, (100, 10))
    sce = SingleCellExperiment(assays={"counts": sparse.csr_matrix(mat)})

    y = polish_dataset(sce)
    assert not sparse.issparse(y.assays["counts"])

    y = polish_dataset(sce, reformat_assay_by_density=1)
    assert sparse.issparse(y.assays["counts"])

    y = polish_dataset(sce, reformat_assay_by_density=None)
    assert sparse.issparse(y.assays["counts"])


def test_polish_dataset_attempts_integer_conversions():
    mat = np.random.poisson(3, (100, 10)).astype(float)
    sce = SingleCellExperiment(assays={"counts": mat})

    y = polish_dataset(sce)
    assert y.assays["counts"].dtype == np.int_

    mat = np.random.poisson(0.1, (100, 10)).astype(float)
    sce = SingleCellExperiment(assays={"counts": sparse.csr_matrix(mat)})

    y = polish_dataset(sce)
    assert sparse.issparse(y.assays["counts"])
    assert y.assays["counts"].dtype == np.int_

    mat = np.random.poisson(3, (100, 10)) * 1.5
    sce = SingleCellExperiment(assays={"counts": mat})

    y = polish_dataset(sce)
    assert y.assays["counts"].dtype == np.float_


def test_polish_dataset_works_with_na_values():
    mat = np.random.poisson(0.1, (100, 10))
    mat = mat.astype(np.float_)
    mat.ravel()[np.random.choice(mat.size, 10, replace=False)] = np.nan
    sce = SingleCellExperiment(assays={"counts": sparse.csr_matrix(mat)})

    y = polish_dataset(sce)
    assert sparse.issparse(y.assays["counts"])
    assert y.assays["counts"].dtype == np.float_


def test_polish_dataset_forbids_highly_nested_altexps():
    mat = np.random.poisson(1, (100, 10))
    sce2 = SingleCellExperiment(assays={"counts": mat[:5] + 2})

    sce1 = SingleCellExperiment(
        assays={"counts": mat[:10] + 1}, alternative_experiments={"rando": sce2}
    )

    sce0 = SingleCellExperiment(
        assays={"counts": mat}, alternative_experiments={"rando": sce1}
    )

    with pytest.raises(Exception):
        polish_dataset(sce0)

    y = polish_dataset(sce0, forbid_nested_altexp=False)
    assert (
        y.get_alternative_experiment_names() == sce0.get_alternative_experiment_names()
    )


def test_polish_existing_dataset():
    sce = fetch_dataset("zeisel-brain-2015", "2023-12-14")

    y = polish_dataset(sce)

    assert y.shape == sce.shape
    assert type(y.assays["counts"]) != type(sce.assays["counts"])



from scrnaseq import fetch_latest_version, list_versions

__author__ = "Jayaram Kancherla"
__copyright__ = "Jayaram Kancherla"
__license__ = "MIT"


def test_list_versions():
    versions = list_versions("romanov-brain-2017")

    assert isinstance(versions, list)
    assert "2023-12-19" in versions


def test_latest_version():
    version = fetch_latest_version("romanov-brain-2017")

    assert isinstance(version, str)
    assert "2023-12-19" == version



import sqlite3
from functools import lru_cache
from typing import Union

import pandas as pd
from gypsum_client import cache_directory, fetch_metadata_database
from gypsum_client.search_metadata import (
    GypsumSearchClause,
    search_metadata_text_filter,
)

from .list_datasets import _format_query_results, _sanitize_query_to_output

__author__ = "Jayaram Kancherla"
__copyright__ = "Jayaram Kancherla"
__license__ = "MIT"


@lru_cache
def search_datasets(
    query: Union[str, GypsumSearchClause],
    cache_dir: str = cache_directory(),
    overwrite: bool = False,
    latest: bool = True,
) -> pd.DataFrame:
    """Search for datasets of interest based on matching text in the associated metadata.

    This is a wrapper around
    :py:func:`~gypsum_client.search_metadata.search_metadata_text`.

    The returned DataFrame contains the usual suspects like the title
    and description for each dataset, the number of rows and columns,
    the organisms and genome builds involved, whether the dataset has
    any pre-computed reduced dimensions, and so on.

    More details can be found in the Bioconductor
    `metadata index <https://github.com/ArtifactDB/bioconductor-metadata-index>`_.

    See Also:
        :py:func:`~scrnaseq.list_datasets.list_datasets`, to list all
        available datasets.

        :py:func:`~gypsum_client.search_metadata.search_metadata_text`,
        to search metadata.

    Examples:

    .. code-block:: python

        res = search_datasets("brain")

        res = search_datasets(define_text_query("Neuro%", partial=True")

        res = search_datasets(define_text_query("10090", field="taxonomy_id")

        res = search_datasets(
            define_text_query("GRCm38", field="genome") &
            (define_text_query("neuro%", partial=True) |
                define_text_query("pancrea%", partial=True))
        )

    Args:
        query:
            The search query string or a gypsum.search.object for
            more complex queries.

        cache_directory:
            Path to cache directory.

        overwrite:
            Whether to overwrite the existing cache.
            Defaults to False.

        latest:
            Whether to fetch only the latest versions of datasets.
            Defaults to True.

    Returns:
        A :py:class:`~pandas.DataFrame` where each row corresponds to
        a dataset, containing various columns of metadata.
        Some columns may be lists to capture 1:many mappings.
    """

    bpath = fetch_metadata_database(cache_dir=cache_dir, overwrite=overwrite)

    where = search_metadata_text_filter(query)
    cond = where["where"]
    params = where["parameters"]

    conn = sqlite3.connect(bpath, check_same_thread=False)
    stmt = "SELECT json_extract(metadata, '$') AS meta, versions.asset AS asset, versions.version AS version, path"
    key_names = ["meta", "asset", "version", "path"]

    if not latest:
        stmt += ", versions.latest AS latest"
        key_names.append("latest")

    stmt += " FROM paths LEFT JOIN versions ON paths.vid = versions.vid WHERE versions.project = 'scRNAseq'"

    if latest:
        stmt += " AND versions.latest = 1"

    if cond:
        stmt += " AND " + " AND ".join(cond)
        cursor = conn.execute(stmt, params)
    else:
        cursor = conn.execute(stmt)

    _qresults = cursor.fetchall()
    conn.close()

    results = _format_query_results(_qresults, key_names)
    return _sanitize_query_to_output(results, latest)



import json
import sqlite3
from functools import lru_cache

import pandas as pd
from gypsum_client import (
    cache_directory,
    fetch_metadata_database,
)

__author__ = "Jayaram Kancherla"
__copyright__ = "Jayaram Kancherla"
__license__ = "MIT"


@lru_cache
def list_datasets(
    cache_dir: str = cache_directory(), overwrite: bool = False, latest: bool = True
) -> pd.DataFrame:
    """List all available datasets.

    Example:

        .. code-block:: python

            datasets = list_datasets()

    Args:
        cache_dir:
            Path to cache directory.

        overwrite:
            Whether to overwrite the database in cache.
            Defaults to False.

        latest:
            Whether to only fetch the latest version of each dataset.
            Defaults to True.

    Returns:
        A :py:class:`~pandas.DataFrame` where each row corresponds to a dataset.
        Each row contains title and description for each dataset,
        the number of rows and columns, the organisms and genome builds involved,
        whether the dataset has any pre-computed reduced dimensions, and so on.
        More details can be found in the
        `Bioconductor metadata schema <https://github.com/ArtifactDB/bioconductor-metadata-index>`_.
    """
    db_path = fetch_metadata_database(cache_dir=cache_dir, overwrite=overwrite)
    conn = sqlite3.connect(db_path, check_same_thread=False)

    stmt = "SELECT json_extract(metadata, '$') AS meta, versions.asset AS asset, versions.version AS version, path"
    key_names = ["meta", "asset", "version", "path"]
    if latest is not True:
        stmt = f"{stmt} versions.latest AS latest"
        key_names.append("latest")

    stmt = f"{stmt} FROM paths LEFT JOIN versions ON paths.vid = versions.vid WHERE versions.project = 'scRNAseq'"
    if latest is True:
        stmt = f"{stmt} AND versions.latest = 1"

    _qresults = conn.execute(stmt).fetchall()
    conn.close()

    results = _format_query_results(_qresults, key_names)

    return _sanitize_query_to_output(results, latest)


def _format_query_results(results: list, key_names: list):
    """Format the results from sqlite as a pandas dataframe.

    Key names must be in the exact same order as the query.
    """
    _out = {}
    for k in key_names:
        _out[k] = []

    for r in results:
        for idx, k in enumerate(key_names):
            _out[k].append(r[idx])

    return _out


def _sanitize_query_to_output(results: list, latest: bool, meta_name: str = "meta"):
    _all_paths = [
        None if "/" not in p else p.rsplit("/", 1)[0] for p in results["path"]
    ]

    df = pd.DataFrame(
        {
            "name": results["asset"],
            "version": results["version"],
            "path": _all_paths,
        }
    )
    if not latest:
        _all_latest = [s == 1 for s in results["latest"]]
        df["latest"] = _all_latest

    _all_metas = [json.loads(s) for s in results[meta_name]]

    df["object"] = _extract_atomic_from_json(
        _all_metas, lambda x: x.get("applications", {}).get("takane", {}).get("type")
    )
    df["title"] = _extract_atomic_from_json(_all_metas, lambda x: x.get("title"))
    df["description"] = _extract_atomic_from_json(_all_metas, lambda x: x.get("title"))
    df["taxonomy_id"] = _extract_charlist_from_json(
        _all_metas, lambda x: x.get("taxonomy_id")
    )
    df["genome"] = _extract_charlist_from_json(_all_metas, lambda x: x.get("genome"))

    df["rows"] = _extract_atomic_from_json(
        _all_metas,
        lambda x: x.get("applications", {})
        .get("takane", {})
        .get("summarized_experiment", {})
        .get("rows"),
    )

    df["columns"] = _extract_atomic_from_json(
        _all_metas,
        lambda x: x.get("applications", {})
        .get("takane", {})
        .get("summarized_experiment", {})
        .get("columns"),
    )

    df["assays"] = _extract_charlist_from_json(
        _all_metas,
        lambda x: x.get("applications", {})
        .get("takane", {})
        .get("summarized_experiment", {})
        .get("assays"),
    )
    df["column_annotations"] = _extract_charlist_from_json(
        _all_metas,
        lambda x: x.get("applications", {})
        .get("takane", {})
        .get("summarized_experiment", {})
        .get("column_annotations"),
    )
    df["reduced_dimensions"] = _extract_charlist_from_json(
        _all_metas,
        lambda x: x.get("applications", {})
        .get("takane", {})
        .get("single_cell_experiment", {})
        .get("reduced_dimensions"),
    )
    df["alternative_experiments"] = _extract_charlist_from_json(
        _all_metas,
        lambda x: x.get("applications", {})
        .get("takane", {})
        .get("single_cell_experiment", {})
        .get("alternative_experiments"),
    )

    df["bioconductor_version"] = _extract_atomic_from_json(
        _all_metas, lambda x: x.get("bioconductor_version")
    )
    df["maintainer_name"] = _extract_atomic_from_json(
        _all_metas, lambda x: x.get("maintainer_name")
    )
    df["maintainer_email"] = _extract_atomic_from_json(
        _all_metas, lambda x: x.get("maintainer_email")
    )

    sources = []
    for meta in _all_metas:
        cursources = meta.get("sources")
        if cursources is None:
            sources.append(pd.DataFrame(columns=["provider", "id", "version"]))
        else:
            sources.append(
                pd.DataFrame(
                    {
                        "provider": [s.get("provider") for s in cursources],
                        "id": [s.get("id") for s in cursources],
                        "version": [s.get("version") for s in cursources],
                    }
                )
            )
    df["sources"] = sources

    return df


def _extract_atomic_from_json(metadata, extract):
    return [
        extract(_meta) if extract(_meta) is not None else None for _meta in metadata
    ]


def _extract_charlist_from_json(metadata, extract):
    return [extract(_meta) if extract(_meta) is not None else [] for _meta in metadata]



from biocframe import BiocFrame
from delayedarray import is_sparse, to_dense_array, to_scipy_sparse_matrix
from dolomite_base import read_object
from singlecellexperiment import SingleCellExperiment
from summarizedexperiment import SummarizedExperiment

__author__ = "Jayaram Kancherla"
__copyright__ = "Jayaram Kancherla"
__license__ = "MIT"


def single_cell_load_object(
    path: str,
    metadata: dict = None,
    scrnaseq_realize_assays: bool = False,
    scrnaseq_realize_reduced_dims: bool = True,
    **kwargs,
):
    """Load a ``SummarizedExperiment`` or ``SingleCellExperiment`` object from a file.

    Args:
        path:
            Path to the dataset.

        metadata:
            Metadata for the dataset.
            Defaults to None.

        scrnaseq_realize_assays:
            Whether to realize assays into memory.
            Defaults to False.

        scrnaseq_realize_reduced_dims:
            Whether to realize reduced dimensions into memory.
            Defaults to True.

        **kwargs:
            Further arguments to pass to
            :py:func:`~dolomite_base.read_object.read_object`.

    Returns:
        A `SingleCellExperiment` or a `SummarizedExperiment` derivative of the object.
    """
    obj = read_object(
        path,
        metadata=metadata,
        scrnaseq_realize_assays=scrnaseq_realize_assays,
        scrnaseq_realize_reduced_dims=scrnaseq_realize_reduced_dims,
        **kwargs,
    )

    if isinstance(obj, SummarizedExperiment):
        if scrnaseq_realize_assays:
            _assays = {}
            for y in obj.get_assay_names():
                _assays[y] = realize_array(obj.assay(y))

            obj = obj.set_assays(_assays)

        if isinstance(obj, SingleCellExperiment):
            if scrnaseq_realize_reduced_dims:
                _red_dims = {}
                for z in obj.get_reduced_dim_names():
                    _red_dims[z] = realize_array(obj.reduced_dim(z))

                obj = obj.set_reduced_dims(_red_dims)

    return obj


def realize_array(x):
    """Realize a `ReloadedArray` into a dense array or sparse matrix.

    Args:
        x:
            `ReloadedArray` object.

    Returns:

        Realized array or matrix.
    """
    from dolomite_matrix import ReloadedArray

    if isinstance(x, ReloadedArray):
        if is_sparse(x):
            x = to_scipy_sparse_matrix(x, "csc")
        else:
            x = to_dense_array(x)

    return x


def format_object_metadata(x) -> dict:
    """Format object related metadata.

    Create object-related metadata to validate against the default
    schema from
    :py:func:`~gypsum_client.fetch_metadata_schema.fetch_metadata_schema`.
    This is intended for downstream package developers who are
    auto-generating metadata documents to be validated by
    :py:func:`~gypsum_client.validate_metadata.validate_metadata`.

    Args:
        x:
            An Python object, typically an instance of a BiocPy class.

    Returns:
        Dictionary containing metadata for the object.
    """
    _meta = {}

    if isinstance(x, SummarizedExperiment):
        _meta["summarized_experiment"] = {
            "rows": x.shape[0],
            "columns": x.shape[1],
            "assays": list(x.get_assay_names()),
            "column_annotations": list(x.get_column_names()),
        }

        if isinstance(x, SingleCellExperiment):
            _meta["single_cell_experiment"] = {
                "reduced_dimensions": list(x.get_reduced_dim_names()),
                "alternative_experiments": list(x.get_alternative_experiment_names()),
            }
    elif isinstance(x, BiocFrame):
        _meta["data_frame"] = {
            "rows": len(x),
            "column_names": list(x.get_column_names()),
        }

    return _meta



import json
import os
import shutil
import tempfile

import anndata as ad
import dolomite_base as dl
import numpy as np
import pandas as pd
import pytest
from dolomite_matrix import ReloadedArray
from scrnaseq import fetch_dataset, save_dataset
from delayedarray import to_dense_array

__author__ = "Jayaram Kancherla"
__copyright__ = "Jayaram Kancherla"
__license__ = "MIT"


def test_save_dataset_sce():
    sce = fetch_dataset("zeisel-brain-2015", "2023-12-14")

    meta = {
        "title": "My dataset forked from ziesel brain",
        "description": "This is a copy of the ziesel",
        "taxonomy_id": ["10090"],  # NCBI ID
        "genome": ["GRCh38"],  # genome build
        "sources": [{"provider": "GEO", "id": "GSE12345"}],
        "maintainer_name": "Shizuka Mogami",
        "maintainer_email": "mogami.shizuka@765pro.com",
    }

    tmp = tempfile.mkdtemp()
    save_dataset(sce, tmp, meta)

    # Load the saved AnnData object
    roundtrip = dl.read_object(tmp)

    assert len(roundtrip.get_column_data()) == len(sce.get_column_data())
    assert isinstance(roundtrip.get_assays()["counts"], ReloadedArray)
    assert isinstance(sce.get_assays()["counts"], ReloadedArray)

    # Load and check the metadata
    with open(os.path.join(tmp, "_bioconductor.json")) as f:
        saved_meta = json.load(f)

    assert saved_meta["bioconductor_version"] == "3.14"  # Placeholder version

    # Test validation failure
    meta["title"] = 1234
    with pytest.raises(Exception):
        save_dataset(sce, tmp, meta)

    shutil.rmtree(tmp)


def test_save_dataset_anndata():
    data = np.random.poisson(1, (10, 100))
    adata = ad.AnnData(data)
    adata.obs["foo"] = np.random.choice(list("abcdefghijklmnopqrstuvwxyz"), 10)
    adata.var_names = [f"GENE_{i+1}" for i in range(adata.n_vars)]
    adata.obs_names = list("ABCDEFGHIJ")
    adata.layers["counts"] = data

    meta = {
        "title": "My dataset forked from ziesel brain",
        "description": "This is a copy of the ziesel",
        "taxonomy_id": ["10090"],  # NCBI ID
        "genome": ["GRCh38"],  # genome build
        "sources": [{"provider": "GEO", "id": "GSE12345"}],
        "maintainer_name": "Shizuka Mogami",
        "maintainer_email": "mogami.shizuka@765pro.com",
    }

    tmp = tempfile.mkdtemp()
    save_dataset(adata, tmp, meta)

    # Load the saved AnnData object
    roundtrip = dl.read_object(tmp)

    assert len(roundtrip.get_column_data()) == 10
    assert isinstance(roundtrip.get_assays()["counts"], ReloadedArray)
    assert isinstance(adata.layers["counts"], np.ndarray)
    assert np.array_equal(
        to_dense_array(roundtrip.get_assays()["counts"]).transpose(),
        adata.layers["counts"],
    )

    # Load and check the metadata
    with open(os.path.join(tmp, "_bioconductor.json")) as f:
        saved_meta = json.load(f)

    assert saved_meta["bioconductor_version"] == "3.14"  # Placeholder version

    # Test validation failure
    meta["title"] = 1234
    with pytest.raises(Exception):
        save_dataset(adata, tmp, meta)

    shutil.rmtree(tmp)



import json
import os
import shutil
from functools import singledispatch
from typing import Any

import dolomite_base as dl
from gypsum_client import fetch_metadata_schema, validate_metadata
from singlecellexperiment import SingleCellExperiment

from .utils import format_object_metadata

__author__ = "Jayaram Kancherla"
__copyright__ = "Jayaram Kancherla"
__license__ = "MIT"


@singledispatch
def save_dataset(x: Any, path, metadata):
    """Save a dataset to disk.

    Args:
        x:
            An object containing single-cell data.
            May be a derivative of SummarizedExperiment or AnnData.

        path:
            Path to a new directory to save the dataset.

        metadata:
            Dictionary containing the metadata for this dataset.
            see the schema returned by
            :py:func:`~gypsum_client.fetch_metadata_schema.fetch_metadata_schema`.

            Note that the ``applications.takane`` property will be automatically
            added by this function and does not have to be supplied.

    See Also:
        `metadata index <https://github.com/ArtifactDB/bioconductor-metadata-index>`_,
        on the expected schema for the metadata.

        :py:func:`~scrnaseq.polish_dataset.polish_dataset`,
        to polish ``x`` before saving it.

        :py:func:`~scrnaseq.upload_dataset.upload_dataset`, to upload the saved contents.

    Example:

        .. code-block:: python

            # Fetch an existing dataset
            # or create your own ``SingleCellExperiment``
            # or ``AnnData`` object.
            sce = scrnaseq.fetch_dataset("zeisel-brain-2015", "2023-12-14")

            # Provide dataset level metadata for search and findability
            meta = {
                "title": "My dataset made from ziesel brain",
                "description": "This is a copy of the ziesel",
                "taxonomy_id": ["10090"], # NCBI ID
                "genome": ["GRCh38"], # genome build
                "sources": [{"provider": "GEO", "id": "GSE12345"}],
                "maintainer_name": "Shizuka Mogami",
                "maintainer_email": "mogami.shizuka@765pro.com",
            }

            import shutil
            import tempfile

            cache_dir = tempfile.mkdtemp()

            # Make sure the directory is clean
            shutil.rmtree(cache_dir)

            # Save the dataset
            scrnaseq.save_dataset(sce, cache_dir, meta)
    """
    raise NotImplementedError(
        f"'save_dataset' is not supported for objects of class: {type(x)}"
    )


@save_dataset.register
def save_dataset_sce(x: SingleCellExperiment, path: str, metadata: dict):
    """Save :py:class:`~singlecellexperiment.SingleCellExperiment.SingleCellExperiment` to disk."""
    schema = fetch_metadata_schema()

    if "bioconductor_version" not in metadata:
        metadata["bioconductor_version"] = "3.14"  # current release

    validate_metadata(metadata, schema)

    if os.path.exists(path):
        shutil.rmtree(path)

    dl.save_object(x, path, reloaded_array_reuse_mode="symlink")

    takane = format_object_metadata(x)
    takane["type"] = dl.read_object_file(path)["type"]

    if "applications" not in metadata:
        metadata["applications"] = {}

    metadata["applications"]["takane"] = takane

    # Second validation with the takane metadata.
    contents = json.dumps(metadata, indent=4)
    validate_metadata(json.loads(contents), schema=schema)
    with open(os.path.join(path, "_bioconductor.json"), "w") as f:
        f.write(contents)

    return


has_anndata = False
try:
    import anndata

    has_anndata = True
except Exception:
    pass

if has_anndata:

    @save_dataset.register
    def save_dataset_anndata(x: anndata.AnnData, path: str, metadata: dict):
        """Save :py:class:`~anndata.AnnData` to disk."""
        _sce = SingleCellExperiment.from_anndata(x)
        return save_dataset(_sce, path, metadata)



import numpy as np
import pytest
import scipy.sparse as sp
from delayedarray import is_sparse
from dolomite_matrix import ReloadedArray
from scrnaseq import fetch_dataset, fetch_metadata
from singlecellexperiment import SingleCellExperiment

__author__ = "Jayaram Kancherla"
__copyright__ = "Jayaram Kancherla"
__license__ = "MIT"


def test_fetch_dataset():
    sce = fetch_dataset("zeisel-brain-2015", "2023-12-14")
    assert isinstance(sce, SingleCellExperiment)

    # Correctly creates ReloadedMatrix objects.
    ass = sce.get_assays()
    assert all(isinstance(a, ReloadedArray) for _, a in ass.items())
    assert all(is_sparse(a) for _, a in ass.items())

    ass_0 = ass["counts"]
    assert "zeisel-brain-2015" in ass_0.seed.path
    assert "2023-12-14" in ass_0.seed.path

    # Works with realization options.
    sce = fetch_dataset("zeisel-brain-2015", "2023-12-14", realize_assays=True)
    ass = sce.get_assays()
    assert all(isinstance(a, (sp.csc_matrix, sp.csr_matrix)) for _, a in ass.items())

    alt_exps = sce.get_alternative_experiments()
    for altname, alt in alt_exps.items():
        alt_exp_ass = alt.get_assays()
        assert all(isinstance(a, (np.ndarray)) for _, a in alt_exp_ass.items())


def test_fetch_dataset_realizes_reduced_dimensions():
    sce = fetch_dataset("aztekin-tail-2019", "2023-12-14", realize_reduced_dims=False)
    red_dim = sce.get_reduced_dims()
    assert all(isinstance(a, ReloadedArray) for _, a in red_dim.items())

    sce = fetch_dataset("aztekin-tail-2019", "2023-12-14", realize_reduced_dims=True)
    red_dim = sce.get_reduced_dims()
    assert all(isinstance(a, np.ndarray) for _, a in red_dim.items())


def test_fetch_metadata():
    meta = fetch_metadata("zeisel-brain-2015", "2023-12-14")
    assert "Brain structure" in meta["title"]
    assert meta["taxonomy_id"][0] == "10090"



from gypsum_client import cache_directory, rest_url, upload_directory

__author__ = "Jayaram Kancherla"
__copyright__ = "Jayaram Kancherla"
__license__ = "MIT"


def upload_dataset(
    directory: str,
    name: str,
    version: str,
    package: str = "scRNAseq",
    cache_dir: str = cache_directory(),
    deduplicate: bool = True,
    probation: bool = False,
    url: str = rest_url(),
    token: str = None,
    concurrent: int = 1,
    abort_failed: bool = True,
):
    """Upload the dataset to the gypsum bucket.

    This is a wrapper around
    :py:func:`~gypsum_client.upload_file_actions.upload_directory`
    specific to the `scRNAseq` package.

    See Also:
        :py:func:`~gypsum_client.upload_file_actions.upload_directory`,
        to upload a directory to the gypsum backend.

    Args:
        Name:
            Dataset name.

        version:
            Version name.

        directory:
            Path to a directory containing the ``files`` to be uploaded.
            This directory is assumed to correspond to a version of an asset.

        cache_dir:
            Path to the cache for saving files, e.g., in
            :py:func:`~gypsum_client.save_operations.save_version`.

            Used to convert symbolic links to upload links,see
            :py:func:`~gypsum_client.prepare_directory_for_upload.prepare_directory_upload`.

        deduplicate:
            Whether the backend should attempt deduplication of ``files``
            in the immediately previous version.
            Defaults to True.

        probation:
            Whether to perform a probational upload.
            Defaults to False.

        url:
            URL of the gypsum REST API.

        token:
            GitHub access token to authenticate to the gypsum REST API.

        concurrent:
            Number of concurrent downloads.
            Defaults to 1.

        abort_failed:
            Whether to abort the upload on any failure.

            Setting this to `False` can be helpful for diagnosing upload problems.

    Returns:
        `True` if successfull, otherwise `False`.
    """
    return upload_directory(
        directory,
        package,
        name,
        version,
        cache_dir=cache_dir,
        deduplicate=deduplicate,
        probation=probation,
        url=url,
        token=token,
        concurrent=concurrent,
        abort_failed=abort_failed,
    )



import json
import os
import shutil
import tempfile

import anndata as ad
import dolomite_base as dl
import dolomite_matrix as dlm
import numpy as np
import pandas as pd
import pytest
from biocframe import BiocFrame
from gypsum_client import prepare_directory_upload
from scipy import sparse as sp
from scrnaseq import fetch_dataset, save_dataset, upload_dataset
from singlecellexperiment import SingleCellExperiment

__author__ = "Jayaram Kancherla"
__copyright__ = "Jayaram Kancherla"
__license__ = "MIT"


def generate_sce():
    mat = np.random.poisson(1, (100, 10))
    row_names = [f"GENE_{i}" for i in range(mat.shape[0])]
    col_names = list("ABCDEFGHIJ")
    sce = SingleCellExperiment(
        assays={"counts": mat},
        row_data=BiocFrame(row_names=row_names),
        column_data=BiocFrame(row_names=col_names),
    )
    return sce


def test_file_listing_without_special_elements():
    sce = generate_sce()

    meta = {
        "title": "My dataset forked from ziesel brain",
        "description": "This is a copy of the ziesel",
        "taxonomy_id": ["10090"],  # NCBI ID
        "genome": ["GRCh38"],  # genome build
        "sources": [{"provider": "GEO", "id": "GSE12345"}],
        "maintainer_name": "Shizuka Mogami",
        "maintainer_email": "mogami.shizuka@765pro.com",
    }

    tmp = tempfile.mkdtemp()
    save_dataset(sce, tmp, meta)

    listing = prepare_directory_upload(tmp, links="always")
    assert len(listing["links"]) == 0
    assert sorted(listing["files"]) == sorted(
        [
            os.path.relpath(os.path.join(dp, f), tmp)
            for dp, dn, filenames in os.walk(tmp)
            for f in filenames
        ]
    )

    shutil.rmtree(tmp)


def test_file_listing_with_reloaded_arrays():
    sce = generate_sce()

    cache = tempfile.mkdtemp()
    tmp0 = os.path.join(cache, "bucket", "scRNAseq", "test", "foo")
    os.makedirs(os.path.dirname(tmp0), exist_ok=True)

    meta = {
        "title": "My dataset forked from ziesel brain",
        "description": "This is a copy of the ziesel",
        "taxonomy_id": ["10090", "9606"],  # NCBI ID
        "genome": ["GRCh38", "GRCm38"],
        "sources": [
            {"provider": "GEO", "id": "GSE12345"},
            {"provider": "DOI", "id": "123asd/231.123"},
        ],
        "maintainer_name": "Kaori Sakuramori",
        "maintainer_email": "sakuramori.kaori@765pro.com",
    }

    save_dataset(sce, tmp0, meta)

    sce2 = sce.copy()
    sce2.assays["reloaded_assay"] = dlm.ReloadedArray(
        path=f"{tmp0}/assays/0", seed=np.random.poisson(1, (100, 10))
    )

    tmp = tempfile.mkdtemp()
    save_dataset(sce2, tmp, meta)
    with pytest.raises(Exception):
        listing = prepare_directory_upload(tmp, links="always")

    listing = prepare_directory_upload(tmp, links="always", cache_dir=cache)
    assert sorted(x["to.path"] for x in listing["links"]) == sorted(
        ["assays/0/array.h5", "assays/0/OBJECT"]
    )
    assert sorted(x["from.path"] for x in listing["links"]) == sorted(
        ["assays/1/array.h5", "assays/1/OBJECT"]
    )
    assert sorted(
        listing["files"] + [x["from.path"] for x in listing["links"]]
    ) == sorted(
        [
            os.path.relpath(os.path.join(dp, f), tmp)
            for dp, dn, filenames in os.walk(tmp)
            for f in filenames
        ]
    )

    shutil.rmtree(tmp)
    shutil.rmtree(cache)


@pytest.mark.skipif(
    "gh_token" not in os.environ, reason="GitHub token not in environment"
)
def test_actual_upload_works_correctly():
    sce = generate_sce()

    gh_token = os.environ.get("gh_token", None)
    if gh_token is None:
        raise ValueError("GitHub token not in environment")

    meta = {
        "title": "My dataset",
        "description": "This is my dataset",
        "taxonomy_id": ["10090"],
        "genome": ["GRCh38"],
        "sources": [{"provider": "GEO", "id": "GSE12345"}],
        "maintainer_name": "Shizuka Mogami",
        "maintainer_email": "mogami.shizuka@765pro.com",
    }

    tmp = tempfile.mkdtemp()
    save_dataset(sce, tmp, meta)

    app_url = "https://gypsum.artifactdb.com"

    version = str(pd.Timestamp.today().date())
    upload_dataset(tmp, "test", version, probation=True, url=app_url, token=gh_token)
    fetch_dataset.cache_clear()  # Clear cache before fetching

    cache = tempfile.mkdtemp()
    roundtrip = fetch_dataset("test", version, cache=cache, url=app_url)

    assert roundtrip.get_column_data().equals(sce.get_column_data())
    assert np.array_equal(roundtrip.assays["counts"], sce.assays["counts"])
    assert isinstance(roundtrip.assay("counts"), type(sce.assay("counts")))

    shutil.rmtree(tmp)
    shutil.rmtree(cache)



import tempfile

import pandas as pd
from scrnaseq import list_datasets

__author__ = "Jayaram Kancherla"
__copyright__ = "Jayaram Kancherla"
__license__ = "MIT"


def test_list_dataset():
    datasets = list_datasets(cache_dir=tempfile.mkdtemp())

    assert isinstance(datasets, pd.DataFrame)
    assert len(datasets) > 80



import atexit
import json
import os

from dolomite_base import alt_read_object, alt_read_object_function
from gypsum_client import cache_directory, save_file, save_version
from summarizedexperiment import SummarizedExperiment

from .utils import single_cell_load_object

__author__ = "Jayaram Kancherla"
__copyright__ = "Jayaram Kancherla"
__license__ = "MIT"


def fetch_dataset(
    name: str,
    version: str,
    path: str = None,
    package: str = "scRNAseq",
    cache_dir: str = cache_directory(),
    overwrite: bool = False,
    realize_assays: bool = False,
    realize_reduced_dims: bool = True,
    **kwargs,
) -> SummarizedExperiment:
    """Fetch a single-cell dataset from the gypsum backend.

    See Also:
        `metadata index <https://github.com/ArtifactDB/bioconductor-metadata-index>`_,
        on the expected schema for the metadata.

        :py:func:`~scrnaseq.save_dataset.save_dataset` and
        :py:func:`~gypsum_client.upload_file_operations.upload_directory`,
        to save and upload a dataset.

        :py:func:`~scrnaseq.list_datasets.list_datasets` and :py:func:`~scrnaseq.list_versions.list_versions`,
        to get possible values for `name` and `version`.

    Example:

        .. code-block:: python

            sce = fetch_dataset("zeisel-brain-2015", "2023-12-14")

    Args:
        name:
            Name of the dataset.

        version:
            Version of the dataset.

        path:
            Path to a subdataset, if name contains multiple datasets.
            Defaults to None.

        package:
            Name of the package.
            Defaults to "scRNAseq".

        cache_dir:
            Path to cache directory.

        overwrite:
            Whether to overwrite existing files.
            Defaults to False.

        realize_assays:
            Whether to realize assays into memory.
            Defaults to False.

        realize_reduced_dims:
            Whether to realize reduced dimensions into memory.
            Defaults to True.

        **kwargs:
            Further arguments to pass to
            :py:func:`~dolomite_base.read_object.read_object`.

    Returns:
        The dataset as a
        :py:class:`~summarizedexperiment.SummarizedExperiment.SummarizedExperiment`
        or one of its subclasses.
    """

    cache_dir = cache_directory(cache_dir)

    version_path = save_version(
        package, name, version, cache_dir=cache_dir, overwrite=overwrite
    )
    obj_path = (
        version_path if path is None else os.path.join(version_path, path.rstrip("/"))
    )

    old = alt_read_object_function(single_cell_load_object)

    def reset_alt_read_func():
        alt_read_object_function(old)

    atexit.register(reset_alt_read_func)
    return alt_read_object(
        obj_path,
        scrnaseq_realize_assays=realize_assays,
        scrnaseq_realize_reduced_dims=realize_reduced_dims,
        **kwargs,
    )


def fetch_metadata(
    name: str,
    version: str,
    path: str = None,
    package: str = "scRNAseq",
    cache_dir: str = cache_directory(),
    overwrite: bool = False,
):
    """Fetch metadata for a dataset from the gypsum backend.

    See Also:
        :py:func:`~.fetch_dataset`,
        to fetch a dataset.

    Example:

    .. code-block:: python

        meta = fetch_metadata("zeisel-brain-2015", "2023-12-14")

    Args:
        name:
            Name of the dataset.

        version:
            Version of the dataset.

        path:
            Path to a subdataset, if name contains multiple datasets.
            Defaults to None.

        package:
            Name of the package.
            Defaults to "scRNAseq".

        cache:
            Path to the cache directory.

        overwrite:
            Whether to overwrite existing files.
            Defaults to False.

    Returns:
        Dictionary containing metadata for the specified dataset.
    """
    remote_path = "_bioconductor.json" if path is None else f"{path}/_bioconductor.json"
    local_path = save_file(
        package, name, version, remote_path, cache_dir=cache_dir, overwrite=overwrite
    )

    with open(local_path, "r") as f:
        metadata = json.load(f)

    return metadata



from typing import List

import gypsum_client as gypc

__author__ = "Jayaram Kancherla"
__copyright__ = "Jayaram Kancherla"
__license__ = "MIT"


def list_versions(name: str) -> List[str]:
    """List all available versions for a dataset.

    Example:

        .. code-block:: python

            versions = list_versions("romanov-brain-2017")

    Args:
        name:
            Name of the dataset.

    Returns:
        A list of version names.
    """
    return gypc.list_versions("scRNAseq", name)


def fetch_latest_version(name: str) -> str:
    """Fetch latest version for a dataset.

    Example:

        .. code-block:: python

            version = fetch_latest_version("romanov-brain-2017")

    Args:
        name:
            Name of the dataset.

    Returns:
        Latest version name.
    """
    return gypc.fetch_latest("scRNAseq", name)



import pandas as pd
from gypsum_client import define_text_query
from scrnaseq import search_datasets

__author__ = "Jayaram Kancherla"
__copyright__ = "Jayaram Kancherla"
__license__ = "MIT"


def test_search_datasets():
    res = search_datasets("brain")
    assert len(res) > 10
    assert isinstance(res, pd.DataFrame)

    res = search_datasets(define_text_query("Neuro%", partial=True))
    assert isinstance(res, pd.DataFrame)
    assert len(res) > 0

    res = search_datasets(define_text_query("10090", field="taxonomy_id"))
    assert isinstance(res, pd.DataFrame)
    assert len(res) > 0

    res = search_datasets(
        define_text_query("GRCm38", field="genome")
        & (
            define_text_query("neuro%", partial=True)
            | define_text_query("pancrea%", partial=True)
        )
    )
    assert isinstance(res, pd.DataFrame)
    assert len(res) > 0



from typing import Type

import numpy as np
from scipy import sparse as sp
from singlecellexperiment import SingleCellExperiment
from summarizedexperiment import SummarizedExperiment

__author__ = "Jayaram Kancherla"
__copyright__ = "Jayaram Kancherla"
__license__ = "MIT"


def polish_dataset(
    x: Type[SummarizedExperiment],
    reformat_assay_by_density: float = 0.3,
    attempt_integer_conversion: bool = True,
    remove_altexp_coldata: bool = True,
    forbid_nested_altexp: bool = True,
) -> Type[SummarizedExperiment]:
    """Optimize dataset for saving.

    Prepare a
    :py:class:`~summarizedexperiment.SummarizedExperiment.SummarizedExperiment` or
    :py:class:`~singlecellexperiment.SingleCellExperiment.SingleCellExperiment`
    to be saved with :py:func:`scrnaseq.save_dataset.save_dataset`.

    This performs minor changes to improve storage efficiency, especially
    with matrices.

    Args:
        x:
            A :py:class:`~summarizedexperiment.SummarizedExperiment.SummarizedExperiment`
            or one of its derivative.

        reformat_assay_by_density:
            Whether to optimize assay formats based on the density of non-zero values.
            Assays with densities above this number are converted to ordinary dense
            arrays (if they are not already), while those with lower densities are
            converted to sparse matrices.

            This can be disabled by setting it to `None`.

        attempt_integer_conversion:
            Whether to convert double-precision assays containing integer values
            to actually have the integer type.

            This can improve efficiency of downstream applications by avoiding
            the need to operate in double precision.

        remove_altexp_coldata:
            Whether column data for alternative experiments should be removed.
            Defaults to `True` as the alternative experiment column data is
            usually redundant compared to the main experiment.

        forbid_nested_altexp:
            Whether nested alternative experiments (i.e., alternative experiments of
            alternative experiments) should be forbidden.

    Returns:
        A modifed object with the same type as ``x``.
    """
    return _polish_dataset(
        x,
        reformat_assay_by_density,
        attempt_integer_conversion,
        remove_altexp_coldata,
        forbid_nested_altexp,
    )


def _polish_dataset(
    x: Type[SummarizedExperiment],
    reformat_assay_by_density: float,
    attempt_integer_conversion: bool,
    remove_altexp_coldata: bool,
    forbid_nested_altexp: bool,
    level: int = 0,
):
    new_assays = {}
    for asyname, asy in x.assays.items():
        if reformat_assay_by_density is not None:
            density = min(np.mean(asy != 0), np.mean(asy != np.nan))
            if density < reformat_assay_by_density:
                if not sp.issparse(asy):
                    asy = sp.csr_matrix(asy)
            else:
                if sp.issparse(asy):
                    asy = asy.toarray()

        if attempt_integer_conversion:
            if np.issubdtype(asy.dtype, np.floating):
                _cast = False
                if sp.issparse(asy):
                    if not np.any(asy.data % 1 != 0):
                        _cast = True
                elif not np.any(asy % 1 != 0):
                    _cast = True

                if _cast is True:
                    asy = asy.astype(np.int_)

        new_assays[asyname] = asy

    x = x.set_assays(new_assays)

    if isinstance(x, SingleCellExperiment):
        if len(x.get_alternative_experiment_names()) > 0:
            if forbid_nested_altexp and level > 0:
                raise ValueError("Nested alternative experiments are forbidden.")

        new_alts = {}
        for altname, altexp in x.alternative_experiments.items():
            if remove_altexp_coldata:
                altexp = altexp.set_column_data(None)

            altexp = _polish_dataset(
                altexp,
                reformat_assay_by_density=reformat_assay_by_density,
                attempt_integer_conversion=attempt_integer_conversion,
                remove_altexp_coldata=remove_altexp_coldata,
                forbid_nested_altexp=forbid_nested_altexp,
                level=level + 1,
            )

            new_alts[altname] = altexp

        x = x.set_alternative_experiments(new_alts)

    return x
